// Code generated by mkcnames -i a.out.go -o cnames.go -p loong64; DO NOT EDIT.

package loong64

// This order should be strictly consistent to that in a.out.go.
var cnames0 = []string{
	"NONE",
	"REG",
	"FREG",
	"FCSRREG",
	"FCCREG",
	"VREG",
	"XREG",
	"ARNG",
	"ELEM",
	"ZCON",
	"SCON",
	"UCON",
	"ADD0CON",
	"AND0CON",
	"ADDCON",
	"ANDCON",
	"LCON",
	"DCON",
	"SACON",
	"LACON",
	"DACON",
	"EXTADDR",
	"BRAN",
	"SAUTO",
	"LAUTO",
	"ZOREG",
	"SOREG",
	"LOREG",
	"ROFF",
	"ADDR",
	"TLS_LE",
	"TLS_IE",
	"GOTADDR",
	"TEXTSIZE",
	"GOK",
	"NCLASS",
}
